#!/usr/bin/env python

# Scan the .TIL file and convert it into a binary tree in memory

import sys
import os
import array
import string
import convTilLoader

class NodeType:

    def __init__(self):
        self.null = 0
        self.keyword = 1
        self.enumConst = 2
        self.identifier = 3
        self.list = 4
        self.int = 5
        self.stringLit = 6

# Read the list of valid .TIL keywords (each preceded by '#')
# Read the list of valid .TIL enumerated constants (preceded by '$')

class TilConstFiles:

    def __init__(self, dirName):
        self.isConstFilesLoaded = False
        keyWordFileName = "KeyWords.lst"
        enumFileName = "EnumConst.lst"
        keyWordFullFileName = os.path.join(dirName, keyWordFileName)
        enumFullFileName = os.path.join(dirName, enumFileName)
        if not os.path.exists(keyWordFullFileName):
            return
        if not os.path.exists(enumFullFileName):
            return
        f = file(keyWordFullFileName, 'r');
        try:
            fileContents = f.read()
        finally:
            f.close()
        self.keyWordList = fileContents.splitlines()
        f = file(enumFullFileName, 'r');
        try:
            fileContents = f.read()
        finally:
            f.close()
        self.enumList = fileContents.splitlines()
        self.isConstFilesLoaded = True

# Node of binary tree:
# Header: 16 bits, type of node
# Left: 32 bits, pointer to first child node (or immediate data)
# Right: 32 bits, pointer to next node

class ListNode:

    def __init__(self, leftVal, rightVal):
        self.header = 0
        self.left = leftVal
        self.right = rightVal

    def setHeader(self, hdrVal):
        self.header = hdrVal

    def setLeft(self, leftVal):
        self.left = leftVal

    def setRight(self, rightVal):
        self.right = rightVal

    def resetListNode(self):
        self.header = 0
        self.left = 0
        self.right = 0

# page: large array of 16-bit integers
# 1. page divided into groups of 5 array elements each
# 2. Each group corresponds to a node in the binary tree
# 3. 1st array element in each group corresponds to the header
# 4. 2nd and 3rd array elements corresponds to the left pointer
# 5. 4th and 5th array elements corresponds to the left pointer
# 6. 1st array element in each pointer is the page no.
# 7. 2nd array element in each pointer is the page array index
#    (points to header of node: 1st array element in group of 5)

class ListPage:

    def __init__(self, emptyPage):
        self.page = array.array('H', emptyPage)
        self.pageLen = len(emptyPage)

    def getPtrVal(self, nodeIdx):
        aWord = self.page[nodeIdx]
        bWord = self.page[nodeIdx + 1]
        return (aWord << 16) | (bWord & 0xFFFF)

    def getPageIdx(self, ptrVal):
        return (ptrVal >> 16) & 0xFFFF

    def getNodeIdx(self, ptrVal):
        return ptrVal & 0xFFFF

    def getListNodeFromPage(self, listNode, nodeIdx):
        listNode.header = self.page[nodeIdx]
        listNode.left = self.getPtrVal(nodeIdx + 1)
        listNode.right = self.getPtrVal(nodeIdx + 3)
    
    def setListNodeToPage(self, listNode, nodeIdx):
        self.page[nodeIdx] = listNode.header
        self.page[nodeIdx + 1] = self.getPageIdx(listNode.left)
        self.page[nodeIdx + 2] = self.getNodeIdx(listNode.left)
        self.page[nodeIdx + 3] = self.getPageIdx(listNode.right)
        self.page[nodeIdx + 4] = self.getNodeIdx(listNode.right)

# pageBook: list of page objects in memory
# pageLen: no. of 16-bit integers in each page

class ListBook:

    def __init__(self, tilCodeList, keyWordList, enumList, debug):
        self.pageLen = 1000
        self.emptyPage = self.pageLen*[0]
        self.firstPage = ListPage(self.emptyPage)
        self.pageBook = [self.firstPage]
        self.nodeSiz = 5
        self.maxPageIdx = 0
        self.maxNodeIdx = -1
        self.maxPageCount = 32768
        self.stack = []
        self.tilCodeList = tilCodeList
        self.keyWordList = keyWordList
        self.enumList = enumList
        self.debug = debug
        self.currListNode = ListNode(0, 0)
        self.rootNode = ListNode(0, 0)
        self.prevNode = ListNode(0, 0)
        # read the .TIL file and convert it into a binary tree:
        self.loadTilFile()

    def resetStack(self):
        self.getListNode(self.rootNode, self.nodeSiz)
        self.prevNodeIdx = 0
        self.currWordIdx = self.nodeSiz
        self.copyListNode(self.rootNode, self.currListNode)
        self.stack = []

    def splitPtrVal(self, ptrVal):
        leftVal = (ptrVal >> 16) & 0xFFFF
        rightVal = ptrVal & 0xFFFF
        return ListNode(leftVal, rightVal)

    def makePtrVal(self, pageIdx, nodeIdx):
        return (pageIdx << 16) | (nodeIdx & 0xFFFF)

    def copyListNode(self, srcListNode, destListNode):
        destListNode.header = srcListNode.header
        destListNode.left = srcListNode.left
        destListNode.right = srcListNode.right

    def getListNode(self, listNode, ptrVal):
        if ptrVal == 0:
            exit
        aListNode = self.splitPtrVal(ptrVal)
        listPage = self.pageBook[aListNode.left]
        listPage.getListNodeFromPage(listNode, aListNode.right)

    def setListNode(self, listNode, ptrVal):
        if ptrVal <= 0:
            exit
        aListNode = self.splitPtrVal(ptrVal)
        listPage = self.pageBook[aListNode.left]
        listPage.setListNodeToPage(listNode, aListNode.right)

    def addListNode(self, listNode):
        result = 0
        status = False
        if self.maxNodeIdx >= 0:
            self.maxNodeIdx = self.maxNodeIdx + self.nodeSiz
        elif self.maxPageIdx > 0:
            self.maxNodeIdx = 0
        else:
            self.maxNodeIdx = self.nodeSiz
        if self.maxNodeIdx >= self.pageLen:
            self.maxPageIdx = self.maxPageIdx + 1
            if self.maxPageIdx > self.maxPageCount:
                return result, status
            self.pageBook.append(ListPage(self.emptyPage))
            self.maxNodeIdx = 0
        listPage = self.pageBook[self.maxPageIdx]
        listPage.setListNodeToPage(listNode, self.maxNodeIdx)
        result = self.makePtrVal(self.maxPageIdx, self.maxNodeIdx)
        status = True
        return result, status

    def getNextWordIdx(self, wordIdx):
        result = 0
        listNode = self.splitPtrVal(wordIdx)
        pageIdx = listNode.left
        nodeIdx = listNode.right
        if nodeIdx < 0:
            nodeIdx = 0
        else:
            nodeIdx = nodeIdx + self.nodeSiz
        if nodeIdx >= self.pageLen:
            pageIdx = pageIdx + 1
            if pageIdx > self.maxPageCount:
                return result
            nodeIdx = 0
        return self.makePtrVal(pageIdx, nodeIdx)

    def printErr(self, msg):
        print "Error:", msg

    def loadTilFile(self):
        self.errToken = ''
        if self.loadTilWordList():
            print "No errors encountered while loading .TIL file."
        else:
            self.errToken = self.errToken.strip()
            errMsg = "Error encountered loading .TIL file!"
            if self.errToken != "":
                errMsg = errMsg + "\nInvalid token: [" + self.errToken
                errMsg = errMsg + "]\non line no. " + str(self.errLineNo)
            print errMsg
            
    def loadTilWordList(self):
        quoteCh = "'"
        escCh = '\\'
        lineNo = 0
        self.tilWordList = []
        self.prevNodeIdx = 0
        self.rootIdx = 0
        self.tilWordListIdx = 0
        for buf in self.tilCodeList:
            lineNo = lineNo + 1
            self.errLineNo = lineNo
            token = ''
            inStringLit = False
            inEscMode = False
            endOfToken = True
            for j in range(len(buf)):
                ch = buf[j]
                charHandled = False
                if not inStringLit:
                    pass
                elif inEscMode:             # process escaped char
                    token = token + ch
                    inEscMode = False
                    charHandled = True

                if charHandled:
                    pass
                elif inStringLit:           # not inEscMode
                    if ch == quoteCh:        # end string lit
                        token = token + ch
                        if not self.wordAdd(token):
                            return False
                        inStringLit = False
                        endOfToken = True
                    elif ch == escCh:        # backslash
                        inEscMode = True
                    else:                   # middle of string lit
                        token = token + ch
                elif ch == quoteCh:          # start string lit
                    inStringLit = True
                    if not endOfToken and not self.wordAdd(token):
                        return False
                    token = ch
                elif ch in set(['(', ')']):
                    if not endOfToken and not self.wordAdd(token):
                        return False
                    token = ch
                    if not self.wordAdd(token):
                        return False
                    token = ''
                    endOfToken = True
                elif ch in set([' ', chr(9)]):  # white space
                    if not endOfToken and not self.wordAdd(token):
                        return False
                    token = ''
                    endOfToken = True
                elif endOfToken:            # start general token
                    token = ch
                    endOfToken = False
                else:                       # middle of token
                    token = token + ch
            if not endOfToken:
                if not self.wordAdd(token):
                    return False
                token = ''
        self.resetStack()
        print 'No. of lines in input =', lineNo
        debugSave = self.debug
        convObj = convTilLoader.ConvClsVarMth(self)
        self.debug = debugSave
        #print 'end convTilLoader: status =', convObj.status
        return convObj.status

    def wordAdd(self, token):
        if token == ')':
            if len(self.stack) == 0:             # too many ')' chars.
                return False
            currListIdx = self.stack.pop()
            self.getListNode(self.prevNode, currListIdx)
            self.prevNodeIdx = currListIdx
        elif token == '(':
            self.currListNode.resetListNode()
            wordTuple = self.wordAddObj(
                self.tilWordList, token, self.currListNode)
            self.tilWordListIdx = wordTuple[0]
            status = wordTuple[1]
            if not status:
                return False
            if self.prevNodeIdx != 0:
                self.prevNode.right = self.tilWordListIdx
                self.setListNode(self.prevNode, self.prevNodeIdx)
            elif len(self.stack) > 0:
                currListIdx = self.stack.pop()
                self.stack.append(currListIdx)
                self.getListNode(self.prevNode, currListIdx)
                self.prevNodeIdx = currListIdx
                self.prevNode.left = self.tilWordListIdx
                self.setListNode(self.prevNode, self.prevNodeIdx)
            else:
                pass
            self.stack.append(self.tilWordListIdx)
            if self.errToken != '':
                return False
            self.prevNodeIdx = 0
        elif self.prevNodeIdx != 0:
            self.currListNode.resetListNode()
            wordTuple = self.wordAddObj(
                self.tilWordList, token, self.currListNode)
            self.tilWordListIdx = wordTuple[0]
            status = wordTuple[1]
            if not status:
                return False
            self.prevNode.right = self.tilWordListIdx
            self.setListNode(self.prevNode, self.prevNodeIdx)
            self.copyListNode(self.currListNode, self.prevNode)
            self.prevNodeIdx = self.tilWordListIdx
        elif len(self.stack) == 0:
            self.currListNode.resetListNode()
            wordTuple = self.wordAddObj(
                self.tilWordList, token, self.currListNode)
            self.tilWordListIdx = wordTuple[0]
            status = wordTuple[1]
            if not status:
                return False
            self.copyListNode(self.currListNode, self.prevNode)
            self.prevNodeIdx = self.tilWordListIdx
            self.rootIdx = self.tilWordListIdx
            #print "RootIdx =", self.rootIdx
        else:
            currListIdx = self.stack.pop()
            self.stack.append(currListIdx)
            self.getListNode(self.prevNode, currListIdx)
            self.prevNodeIdx = currListIdx
            self.currListNode.resetListNode()
            wordTuple = self.wordAddObj(
                self.tilWordList, token, self.currListNode)
            self.tilWordListIdx = wordTuple[0]
            status = wordTuple[1]
            if not status:
                return False
            self.prevNode.left = self.tilWordListIdx
            self.setListNode(self.prevNode, self.prevNodeIdx)
            self.copyListNode(self.currListNode, self.prevNode)
            self.prevNodeIdx = self.tilWordListIdx
        return True            

    def wordAddObj(self, aList, token, listNode):
        status = False
        result = 0
        token = token.strip()
        if token == '':
            return result, status
        nt = NodeType()
        ch = token[0]
        if ch == '#':
            nodeType = nt.keyword
        elif ch == '$':
            nodeType = nt.enumConst
        elif ch == "'":
            nodeType = nt.stringLit
        elif ch == '_':
            nodeType = nt.identifier
        elif ch == '(':
            nodeType = nt.list
        else:
            nodeType = nt.int
        if nodeType == nt.keyword:
            nodeVal = self.getKeyWordVal(token)
        elif nodeType == nt.enumConst:
            nodeVal = self.getEnumVal(token)
        elif nodeType == nt.stringLit:
            nodeVal = 0
        elif nodeType == nt.identifier:
            nodeVal = self.getIdentifierVal(token)
        elif nodeType == nt.list:
            nodeVal = self.getListVal(token)
        elif nodeType == nt.int:
            nodeVal = self.getIntVal(token)
        else:
            nodeVal = -1
        if nodeVal < 0:
            self.errToken = token
            return result, status
        if (nodeType == nt.identifier) or (nodeType == nt.stringLit):
            aList.append(token)
            nodeVal = len(aList) - 1
        resultTuple = self.addListNode(listNode)
        result = resultTuple[0]
        status = resultTuple[1]
        if not status:
            self.printErr('MasterPage Table Overflow')
            return result, status
        if nodeType == nt.int:
            if ch == '-':
                nodeVal = -nodeVal
        elif nodeType == nt.keyword:
            if nodeVal == 0:
                nodeType = nt.null
        listNode.left = nodeVal
        listNode.header = nodeType
        self.setListNode(listNode, result)
        status = True
        return result, status

    def getKeyWordVal(self, token):
        token = token[1:]
        if token not in self.keyWordList:
            return -1
        for i in range(len(self.keyWordList)):
            if token == self.keyWordList[i]:
                return i
        
    def getEnumVal(self, token):
        token = token[1:]
        if token not in self.enumList:
            return -1
        for i in range(len(self.enumList)):
            if token == self.enumList[i]:
                return i

    def getListVal(self, token):
        if token == '(':
            return 0
        else:
            return -1

    def getIntVal(self, token):
        if token[0] == '-':
            token = token[1:]
        try:
            result = int(token)
        except ValueError:
            result = -1
        return result

    def getIdentifierVal(self, token):
        result = -1
        token = token[1:]
        if token == '':
            return result
        alphaFound = False
        alphaNum = False
        isHyphen = False
        for i in range(len(token)):
            ch = token[i]
            if ch == '_':
                if isHyphen:
                    return result
                alphaNum = False
                isHyphen = False
            elif ch == '-':
                if not alphaNum:
                    return result
                alphaNum = False
                isHyphen = True
            elif ch in string.ascii_letters:
                alphaFound = True
                alphaNum = True
                isHyphen = False
            elif ch not in string.digits:   # invalid char.
                return result
            elif not alphaFound:            # can't start w/ digit
                return result
            else:
                alphaNum = True
                isHyphen = False
        if not alphaFound:                  # must contain letter(s)
            return result
        if isHyphen:                    # can't end w/ hyphen
            return result
        return 0

    def atomic(self, listNode):
        nt = NodeType()
        hdrVal = listNode.header & 0x000F
        return hdrVal != nt.list

    def leftToToken(self, listNode):
        hdrType = listNode.header & 0x000F
        leftVal = listNode.left
        nt = NodeType()
        if hdrType == nt.null:
            return '#nil'
        if hdrType == nt.list:
            return '('
        if hdrType == nt.int:
            return '%d' % leftVal
        if (hdrType == nt.identifier) or (hdrType == nt.stringLit):
            return self.tilWordList[leftVal]
        if hdrType == nt.keyword:
            return '#' + self.keyWordList[leftVal]
        if hdrType == nt.enumConst:
            return '$' + self.enumList[leftVal]
        return ''

    def goParentToken(self):
        if not self.goParent():
            return ''
        return self.leftToToken(self.currListNode)

    def goParent(self):
        if self.stack == []:
            return False
        self.currWordIdx = self.stack.pop()
        self.getListNode(self.currListNode, self.currWordIdx)
        return True
    
    def goChildToken(self):
        if not self.goChild():
            return ''
        return self.leftToToken(self.currListNode)

    def goChild(self):
        if self.atomic(self.currListNode) or (self.currListNode.left == 0):
            return False
        self.stack.append(self.currWordIdx)
        self.currWordIdx = self.currListNode.left
        self.getListNode(self.currListNode, self.currWordIdx)
        return True

    def goNextSiblingToken(self):
        if not self.goNextSibling():
            return ''
        return self.leftToToken(self.currListNode)

    def goNextSibling(self):
        if self.currListNode.right == 0:
            return False
        self.currWordIdx = self.currListNode.right
        self.getListNode(self.currListNode, self.currWordIdx)
        return True

    def goPrevSiblingToken(self):
        if not self.goPrevSibling():
            return ''
        return self.leftToToken(self.currListNode)

    def goPrevSibling(self):        
        if self.stack == []:
            wordIdx = self.rootIdx
        else:
            wordIdx = self.stack.pop()
            self.stack.append(wordIdx)
            self.getListNode(self.prevNode, wordIdx)
            self.prevNodeIdx = wordIdx
            wordIdx = self.prevNode.left
        oldWordIdx = wordIdx
        while wordIdx != self.currWordIdx:
            oldWordIdx = wordIdx
            self.getListNode(self.prevNode, wordIdx)
            if self.prevNodeIdx == 0:
                return False
            self.prevNodeIdx = wordIdx
            if self.prevNode.right == 0:
                self.setListNode(self.prevNode, self.prevNodeIdx)
                return False
            wordIdx = self.prevNode.right
        self.currWordIdx = oldWordIdx
        self.getListNode(self.currListNode, self.currWordIdx)
        return True
    
